@@ -1,6 +1,7 @@
 from abc import ABC, abstractmethod
+from enum import IntEnum, auto
 from types import SimpleNamespace
-from typing import Union, List
+from typing import Union, List, Dict
 import re
 
 
@@ -11,6 +12,22 @@ class Directive:
         self.name = name
 
 
+SPHINX_RULES: List[Directive] = [
+    Directive(
+        pattern=r':(func|meth|class):`\.?(?P<name>[^`]+?)`',
+        replacement=r'`\g<name>`'
+    ),
+    Directive(
+        pattern=r'^:param (?P<param>\S+):',
+        replacement=r'- `\g<param>`:'
+    ),
+    Directive(
+        pattern=r'^:return:',
+        replacement=r'Returns:'
+    )
+]
+
+
 RST_DIRECTIVES: List[Directive] = [
     Directive(
         pattern=r'\.\. versionchanged:: (?P<version>\S+)(?P<end>$|\n)',
@@ -25,8 +42,8 @@ RST_DIRECTIVES: List[Directive] = [
         replacement=r'*Deprecated since \g<version>*\g<end>'
     ),
     Directive(
-        pattern=r'\.\. warning::/g',
-        replacement=r'**Warning**'
+        pattern=r'\.\. warning::',
+        replacement=r'**Warning**:'
     ),
     Directive(
         pattern=r'\.\. seealso::(?P<short_form>.*)(?P<end>$|\n)',
@@ -50,9 +67,12 @@ RST_DIRECTIVES: List[Directive] = [
     ),
     Directive(
         pattern=r':math:`(?P<latex>[^`]+?)`',
-        # this will give $latex$, the second dollar is an escape character
         replacement=r'$\g<latex>$'
     ),
+    Directive(
+        pattern=r'\.\. math:: (?P<latex>[^`]+?)(?P<end>$|\n)',
+        replacement=r'$$\g<latex>$$\g<end>'
+    ),
     Directive(
         pattern=r'\.\. highlight:: (?P<language>.+)(?P<end>$|\n)',
         replacement=r'',
@@ -62,6 +82,48 @@ RST_DIRECTIVES: List[Directive] = [
         pattern=r'\.\. (code-block|productionlist)::(?P<language>.*)(?P<end>$|\n)',
         replacement=r'\g<end>',
         name='code-block'
+    ),
+    *SPHINX_RULES
+]
+
+
+_RST_SECTIONS = {
+    'Parameters',
+    'Returns',
+    'See Also',
+    'Examples',
+    'Attributes',
+    'Notes',
+    'References'
+}
+
+
+# TODO: type with RstSection = Literal[], and generate _RST_SECTIONS out of it once
+#  support for Python 3.6 can be safely dropped
+SECTION_DIRECTIVES: Dict[str, List[Directive]] = {
+    'Parameters': [
+        Directive(
+            pattern=r'^(?P<other_args>\*\*kwargs|\*args)$',
+            replacement=r'- `\g<other_args>`'
+        ),
+        Directive(
+            pattern=r'^(?P<arg1>[^:\s]+\d), (?P<arg2>[^:\s]+\d), \.\.\. : (?P<type>.+)$',
+            replacement=r'- `\g<arg1>`, `\g<arg2>`, `...`: \g<type>'
+        )
+    ],
+    'References': [
+        Directive(
+            pattern=r'^\.\. \[(?P<number>\d+)\] (?P<first_line>.+)$',
+            replacement=r' - [\g<number>] \g<first_line>'
+        )
+    ]
+}
+
+
+ESCAPING_RULES: List[Directive] = [
+    Directive(
+        pattern=r'__(?P<text>\S+)__',
+        replacement=r'\_\_\g<text>\_\_'
     )
 ]
 
@@ -76,16 +138,6 @@ def _find_directive_pattern(name: str):
 HIGHLIGHT_PATTERN = _find_directive_pattern('highlight')
 CODE_BLOCK_PATTERN = _find_directive_pattern('code-block')
 
-_RST_SECTIONS = {
-    'Parameters',
-    'Returns',
-    'See Also',
-    'Examples',
-    'Attributes',
-    'Notes',
-    'References'
-}
-
 
 def looks_like_rst(value: str) -> bool:
     # check if any of the characteristic sections (and the properly formatted underline) is there
@@ -96,7 +148,7 @@ def looks_like_rst(value: str) -> bool:
         if re.search(directive.pattern, value):
             return True
     # allow "text::" or "text ::" but not "^::$" or "^:::$"
-    return re.search(r'(\s|\w)::\n', value) or '\n>>> ' in value
+    return bool(re.search(r'(\s|\w)::\n', value) or '\n>>> ' in value)
 
 
 class IBlockBeginning(SimpleNamespace):
@@ -137,9 +189,98 @@ class IParser(ABC):
     follower: Union['IParser', None] = None
 
 
+class TableParser(IParser):
+
+    class State(IntEnum):
+        AWAITS = auto()
+        PARSING_HEADER = auto()
+        PARSED_HEADER = auto()
+        PARSING_ROWS = auto()
+        FINISHED = auto()
+
+    outer_border_pattern = r'^=+( +=+)+$'
+
+    _state: int
+    _column_starts: List[int]
+    _columns: List[str]
+    _rows: List[List[str]]
+    _max_sizes: List[int]
+
+    def _reset_state(self):
+        self._state = TableParser.State.AWAITS
+        self._column_starts = []
+        self._columns = []
+        self._rows = []
+        self._max_sizes = []
+
+    def can_parse(self, line: str) -> bool:
+        return bool(re.match(self.outer_border_pattern, line))
+
+    def initiate_parsing(self, line: str, current_language: str) -> IBlockBeginning:
+        self._reset_state()
+        match = re.match(self.outer_border_pattern, line)
+        assert match
+        self._column_starts = []
+        previous = ' '
+        for i, char in enumerate(line):
+            if char == '=' and previous == ' ':
+                self._column_starts.append(i)
+            previous = char
+        self._max_sizes = [0 for i in self._column_starts]
+        self._state = TableParser.State.PARSING_HEADER
+        return IBlockBeginning(remainder='')
+
+    def can_consume(self, line: str) -> bool:
+        return bool(self._state != TableParser.State.FINISHED)
+
+    def consume(self, line: str) -> None:
+        states = TableParser.State
+        if self._state == states.PARSING_HEADER:
+            self._columns = self._split(line)
+            self._state += 1
+        elif self._state == states.PARSED_HEADER:
+            # TODO: check integrity?
+            self._state += 1
+        elif self._state == states.PARSING_ROWS:
+            match = re.match(self.outer_border_pattern, line)
+            if match:
+                self._state += 1
+            else:
+                self._rows.append(self._split(line))
+
+    def _split(self, line: str) -> List[str]:
+        assert self._column_starts
+        fragments = []
+        for i, start in enumerate(self._column_starts):
+            end = self._column_starts[i + 1] if i < len(self._column_starts) - 1 else None
+            fragment = line[start:end].strip()
+            self._max_sizes[i] = max(self._max_sizes[i], len(fragment))
+            fragments.append(fragment)
+        return fragments
+
+    def _wrap(self, row: List[str], align=str.ljust) -> str:
+        padded_row = [
+            align(e, self._max_sizes[i])
+            for i, e in enumerate(row)
+        ]
+        return '| ' + (' | '.join(padded_row)) + ' |\n'
+
+    def finish_consumption(self, final: bool) -> str:
+        result = self._wrap(self._columns, align=str.center)
+        result += self._wrap([
+            '-' * size
+            for size in self._max_sizes
+        ])
+
+        for row in self._rows:
+            result += self._wrap(row)
+
+        return result
+
+
 class BlockParser(IParser):
     enclosure = '```'
-    follower = None
+    follower: Union['IParser', None] = None
     _buffer: List[str]
     _block_started: bool
 
@@ -157,10 +298,10 @@ class BlockParser(IParser):
 
     def consume(self, line: str):
         if not self._block_started:
-            raise ValueError('Block has not started')
+            raise ValueError('Block has not started')   # pragma: no cover
         self._buffer.append(line)
 
-    def finish_consumption(self, final: bool):
+    def finish_consumption(self, final: bool) -> str:
         # if the last line is empty (e.g. a separator of intended block), discard it
         if self._buffer[len(self._buffer) - 1].strip() == '':
             self._buffer.pop()
@@ -189,7 +330,7 @@ class IndentedBlockParser(BlockParser, ABC):
     def can_consume(self, line: str) -> bool:
         if self._is_block_beginning and line.strip() == '':
             return True
-        return (len(line) > 0 and re.match(r'^\s', line[0])) or len(line) == 0
+        return bool((len(line) > 0 and re.match(r'^\s', line[0])) or len(line) == 0)
 
     def consume(self, line: str):
         if self._is_block_beginning:
@@ -289,6 +430,8 @@ class ExplicitCodeBlockParser(IndentedBlockParser):
 
     def initiate_parsing(self, line: str, current_language: str) -> IBlockBeginning:
         match = re.match(CODE_BLOCK_PATTERN, line)
+        # already checked in can_parse
+        assert match
         self._start_block(match.group('language').strip() or current_language)
         return IBlockBeginning(remainder='')
 
@@ -298,7 +441,8 @@ BLOCK_PARSERS = [
     NoteBlockParser(),
     MathBlockParser(),
     ExplicitCodeBlockParser(),
-    DoubleColonBlockParser()
+    DoubleColonBlockParser(),
+    TableParser()
 ]
 
 RST_SECTIONS = {
@@ -306,8 +450,13 @@ RST_SECTIONS = {
     for section in _RST_SECTIONS
 }
 
+DIRECTIVES = [
+    *RST_DIRECTIVES,
+    *ESCAPING_RULES
+]
 
-def rst_to_markdown(text: str):
+
+def rst_to_markdown(text: str) -> str:
     """
     Try to parse docstrings in following formats to markdown:
     - https://www.python.org/dev/peps/pep-0287/
@@ -342,7 +491,7 @@ def rst_to_markdown(text: str):
         nonlocal lines_buffer
         lines = '\n'.join(lines_buffer)
         # rst markup handling
-        for directive in RST_DIRECTIVES:
+        for directive in DIRECTIVES:
             lines = re.sub(directive.pattern, directive.replacement, lines)
 
         for (section, header) in RST_SECTIONS.items():
@@ -353,7 +502,7 @@ def rst_to_markdown(text: str):
 
     for line in text.split('\n'):
         if is_first_line:
-            signature_match = re.match(r'^(?P<name>\w+)\((?P<params>.*)\)$', line)
+            signature_match = re.match(r'^(?P<name>\S+)\((?P<params>.*)\)$', line)
             if signature_match and signature_match.group('name').isidentifier():
                 markdown += '```python\n' + line + '\n```\n'
                 continue
@@ -384,17 +533,19 @@ def rst_to_markdown(text: str):
 
             # ok, we are not in any code block (it may start with the next line, but this line is clear - or empty)
 
-            if trimmed_line.rstrip() in RST_SECTIONS:
-                most_recent_section = trimmed_line.rstrip()
-
             # lists handling: items detection
-            match = re.match(r'^(?P<argument>[^: ]+) : (?P<type>.+)$', trimmed_line)
+            # this one does NOT allow spaces on the left hand side (to avoid false positive matches)
+            match = re.match(r'^(?P<argument>[^:\s]+) : (?P<type>.+)$', trimmed_line)
             if match:
                 line = '- `' + match.group('argument') + '`: ' + match.group('type') + ''
-            elif most_recent_section == 'Parameters':
-                kwargs_or_args_match = re.match(r'^(?P<other_args>\*\*kwargs|\*args)$', trimmed_line)
-                if kwargs_or_args_match:
-                    line = '- `' + kwargs_or_args_match.group('other_args') + '`'
+            else:
+                if most_recent_section in SECTION_DIRECTIVES:
+                    for section_directive in SECTION_DIRECTIVES[most_recent_section]:
+                        if re.match(section_directive.pattern, trimmed_line):
+                            line = re.sub(section_directive.pattern, section_directive.replacement, trimmed_line)
+                            break
+                if trimmed_line.rstrip() in RST_SECTIONS:
+                    most_recent_section = trimmed_line.rstrip()
 
             # change highlight language if requested
             # this should not conflict with the parsers starting above
